module Main where


import System.Environment(getArgs, getProgName)
import System.Exit(exitFailure, exitSuccess)
import Data.List(sortOn)
import Data.List.Grouping(splitEvery)
import System.Random(getStdGen, randomRs)
import Control.Monad(when)
import Control.Parallel(par, pseq)
import Control.Parallel.Strategies(parMap, rdeepseq)
import qualified Data.Map as M

import GA
import NQueens
import TSP

main :: IO ()
main =  do
  args <- getArgs
  case args of
    iters:popSize:mutationPct:pArgs -> do
      -- do some argparsing here
      when (read popSize <= (0::Integer)) (do
          putStrLn "Population size must be larger than 0."
          exitFailure)

      case pArgs of
        ["nqueens", n] -> do
          let d = NQueens $ read n
          geneticAlgorithm d (read iters) (read popSize) (read mutationPct)
        ["tsp", n'] -> do
          let n = read n'
              cm = M.fromList [(i, M.fromList [(j, i+j+1) | j <- [0..(n-1)]])| i <- [0..(n-1)]]
              d = TSP n cm
          geneticAlgorithm d (read iters) (read popSize) (read mutationPct)
        _ -> do
          putStrLn "Unimplemented problem-string or poorly written problem-args"
          exitFailure

    _ -> do
      prog <- getProgName
      putStrLn $ "Usage: "++prog++" <num-iter> <pop-size> <mutation-pct> <problem-string> …<problem-args>"
      exitFailure


merge :: [(Int, [Int])] -> [(Int, [Int])] -> [(Int, [Int])]
merge l [] = l
merge [] l = l
merge l1@(x1@(f1, _):t1) l2@(x2@(f2, _):t2)
  | f1 > f2   = x1: merge t1 l2
  | otherwise = x2: merge l1 t2

-- depth limited sorting
parMergeSort :: Int -> [(Int, [Int])] -> [(Int, [Int])]
parMergeSort _ [] = []
parMergeSort _ l@[_] = l
parMergeSort d l
  | d == 0    = sortOn (negate . fst) l
  | otherwise = mergedFirst `par` mergedSecond `pseq` merge mergedFirst mergedSecond
  where
    mergedFirst = parMergeSort (d-1) first
    mergedSecond = parMergeSort (d-1) second
    (first, second) = splitAt halfway l
    (halfway, _) = quotRem (length l) 2


geneticAlgorithm :: GAData a => a -> Int -> Int -> Float -> IO ()
geneticAlgorithm d iters popSize mutationPct = do
  g <- getStdGen
  let randOdds = randomRs (0.0 :: Float, 1.0) g
      randVals = randomRs (dataRange d) g
      (initialPop, newRandomVals) = initialize d randVals popSize

  -- we want to loop with this initial population until done
  gaLoop initialPop randOdds newRandomVals iters
      where
          gaLoop pop rs rvs curIters = do

            -- 1) Get the fitness of whole population in order. DeepSeq for nonsequential eval
            let fitPop = parMap rdeepseq (\x -> (fitness d x, x)) pop
                sortedFitPop = parMergeSort 7 fitPop  -- this could be a parameter, but limited by our machines' cores
                (topFitness, topSolution):_ = sortedFitPop

            -- Exit if optimal
            when (topFitness == maxFitness d) (do
                putStrLn $ "Optimal solution found:\n" ++ showSol d topSolution
                exitSuccess)

            -- 2) Get the proportional probabilities for each chromosome
            let probs = map (\(f, _) -> (fromIntegral f :: Float) / fromIntegral (maxFitness d)) sortedFitPop
                totalProb = sum probs

                -- 3) take the number of probabilities needed to select elements for crossover
                (crossRs, rs1) = splitAt (length pop * 2) rs
                -- this is probably a bottleneck
                pairList = map (\r -> pickrandom (r*totalProb) sortedFitPop probs 0) crossRs
                pairs  = map (\[x, y] -> (x,y)) (splitEvery 2 pairList)
                crossed = parMap rdeepseq (crossover d) pairs

                numRs = length pop * repLength d

                spltAtEvery l = (splitEvery (repLength d) old , new)
                  where (old, new) = splitAt numRs l

                -- 4) take odds and elements to switch to when mutating randomly
                (splitRs, rs2) = spltAtEvery rs1
                (splitRvs, nextRvs) = spltAtEvery rvs
                zippedList = zip3 splitRs splitRvs crossed

                mutated = parMap rdeepseq (\(rands, rvals, sol) -> mutate d mutationPct rands rvals sol) zippedList

            -- VERBOSE
            -- putStrLn $ "The top rated solution for this generation is:\n" ++ showSol t topSolution ++ "\nwith fitness " ++ show topFitness
            -- putStrLn $ "The top fitness is " ++ show topFitness
            if curIters > 1
               then gaLoop mutated rs2 nextRvs (curIters-1)
               else do
                 putStrLn $ "The top rated solution after all iterations is:\n" ++ showSol d topSolution ++ "\nwith fitness " ++ show topFitness

            where
                -- take first element that exceeds a cumulative prob threshold
                pickrandom _ [(_,c)] _ _ = c
                pickrandom r ((_,c):ct) (p:pt) upto
                  | upto + p >= r = c
                  | otherwise = pickrandom r ct pt (upto + p)
                pickrandom _ _ _ _ = []
